@JS()
library web_components;

import 'dart:async';
import 'dart:html';

import 'package:js/js.dart';

bool _customElementsEnabled = false;

void enableCustomElements() {
  if (!_customElementsEnabled) {
    eval('''
      class CustomElement extends HTMLElement {
          constructor() {
              super();
              this.component = null;
          }
      
          connectedCallback() {
              this.component.connectedCallback();
          }
      
          disconnectedCallback() {
              this.component.disconnected();
          }
      
          nativeAttachShadow(mode, delegatesFocus) {
              return this.attachShadow({mode: mode, delegatesFocus: delegatesFocus});
          }
      }
      window.CustomElement = CustomElement;
      
      function nativeDefineElement(name, construct) {
          customElements.define(name, class extends CustomElement {
              constructor() {
                  super();
                  construct(this);
              }
          });
      }
    ''');
    _customElementsEnabled = true;
  }
}

@JS('eval')
external void eval(String javascript);

@JS('CustomElement')
abstract class JSCustomElement<T extends CustomElement> implements HtmlElement {
  T component;

  @JS('nativeAttachShadow')
  external ShadowRoot nativeAttachShadow(String mode, bool delegatesFocus);
}

typedef CustomElementConstructor = CustomElement Function(JSCustomElement e);

void defineElement(String name, CustomElementConstructor constructor) {
  nativeDefineElement(name, allowInterop(constructor));
}

@JS('nativeDefineElement')
external void nativeDefineElement(String name, CustomElementConstructor constructor);

class CustomElement {
  static final List<CustomElement> _elementStack = [];
  final JSCustomElement element;

  CustomElement(this.element) {
    element.component = this;
  }

  void connectedCallback() {
    connected();
    _elementStack.add(this);
    Timer(Duration(seconds: 0), resolveChildren);
  }

  void resolveChildren() {
    while (_elementStack.isNotEmpty) {
      var element = _elementStack.removeLast();
      element.childrenConnected();
    }
  }

  void connected() {}

  void childrenConnected() {}

  void disconnected() {}

  ShadowRoot attachShadow({bool open = true, bool delegatesFocus = false}) =>
      element.nativeAttachShadow(open ? 'open' : 'closed', delegatesFocus);

  T parentOfType<T>() {
    var parent = element.parentNode;
    if (parent is JSCustomElement) {
      if (parent.component is T) {
        return parent.component as T;
      }
    }
    throw 'No parent found of type $T';
  }
}
